#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include "gfx.h"
#include "lodsav.h"
#include "pngio.h"

#define PIXSIZ 24

#define iW  (32*PIXSIZ+33)
#define iH  (32*PIXSIZ+33)
static unsigned int image[iW*iH];
static unsigned char pix[32*32];
static int pW, pH, pixOFFx, pixOFFy;
static pixiw, pixih; 
static unsigned int sma[32*32];
static int color;

static int modified= 0;
static int patmodified= 0;
static char *cfn;
static int npats, spats, cpat;
static char **pats;

static char *newpat(int w,int h)
{
  char *R;
  R= malloc(4+w*h+1);
  sprintf(R,"%02d%02d",w,h);
  memset(R+4, '0', w*h);
  R[4+w*h]= 0;
  return R;
}

static void destroy_pats()
{
  int i;
  if (pats) 
  {
    for(i=0;i<npats;i++) free(pats[i]);
    free(pats);
  }
}

static void mkinitpats(int pp)
{
  spats= 100;
  npats= 1;
  pats= malloc(sizeof(pats[0])*spats);
  pats[0]= newpat(32,32);
  if (pp)
  {
    pats[0][4+11+11*32]= '1';
    pats[0][4+19+11*32]= '2';
    pats[0][4+19+19*32]= '3';
    pats[0][4+11+19*32]= '4';
  }
  cpat= 0;
  modified= 0;
  patmodified= 0;
  cfn= NULL;
}
static int pixXimg(int x)
{
  return 1 + x*(PIXSIZ+1) + pixOFFx;
}

static int pixYimg(int y)
{
  return 1+ y*(PIXSIZ+1) + pixOFFy;
}

static void draw_vert_gridline(int x, unsigned int color)
{
  int y;
  x= pixXimg(x)-1;
  for(y=0;y<pixih;y++)
    image[x+(y+pixOFFy)*iW]= color;
}

static void draw_horz_gridline(int y,unsigned int color)
{
  int x;
  y= pixYimg(y)-1;
  for(x=0;x<pixiw;x++)
    image[(x+pixOFFx)+y*iW]= color;
}
static unsigned int colors[]=
  { 0,  0xFFff0000, 0xFF00ffff, 0xFFeeD5b7, 0xFFee9a00 };

static void render_pix(int ux,int uy,int updsma)
{
  unsigned int col;
  int x,y,dx,dy;
  int smax, smay;
  if (ux<0 || uy<0 || ux>=pW || uy>=pH) return ;
  col= colors[pix[ux+uy*pW]];
  x= pixXimg(ux);
  y= pixYimg(uy);
  for(dx=0;dx<PIXSIZ;dx++)
    for(dy=0;dy<PIXSIZ;dy++)
      image[(dx+x)+(dy+y)*iW]= col;
  gfx_canvas_put(Wcan,x,y,PIXSIZ,PIXSIZ,image,iW,iH,x,y);
  smax= (32-pW)/2+ux;
  smay= (32-pH)/2+uy;
  sma[smax+smay*32]= col;
  if (!updsma) return ;
  gfx_canvas_put(Wsma,smax,smay, 1,1, sma, 32, 32, smax, smay);
}

static void render_bg()
{
  int i,x,y;
  for(i=0;i<iW*iH;i++) image[i]= 0x333300;
  for(y=0;y<pixih;y++)
   for(x=0;x<pixiw;x++)
    image[(x+pixOFFx)+(y+pixOFFy)*iW]= 0x00;
  for(x=0;x<=pW;x++)
    draw_vert_gridline(x, x%5==0 ? 0x444444 : 0x222222);
  for(y=0;y<=pH;y++)
    draw_horz_gridline(y, y%5==0 ? 0x444444 : 0x222222);
  gfx_canvas_put(Wcan,0,0,iW,iH,image,iW,iH,0,0);
}

static void render_all_pix()
{
  int i;
  if (!pW) return ;
  memset(sma,0,sizeof(sma));
  for(i=0;i<pW*pH;i++)
   if (pix[i]) render_pix(i%pW,i/pW,0);
  gfx_canvas_put(Wsma, 0,0,32,32,sma,32,32,0,0);
}


static void setsize(int w,int h,int cp)
{
  static unsigned char tempix[32*32];
  int x,y,tW,tH;
  if (w<=0 || h<=0 || w>32 || h>32) return ;
  pixiw= w*(PIXSIZ+1)+1;
  pixih= h*(PIXSIZ+1)+1;
  pixOFFx= (iW - pixiw) / 2;
  pixOFFy= (iW - pixih) / 2;
  tW= pW; tH= pH; 
  pW= w; pH=h;
  if (cp)
  {
    memcpy(tempix, pix, sizeof(pix));
    memset(pix, 0, sizeof(pix));
    for(y=0;y<tH && y<pH;y++)
      for(x=0;x<tW && x<pW;x++)
        pix[x+y*pW]= tempix[x+y*tW];
  }
  render_bg();
  if (cp)
    render_all_pix();
}

static void setcolor(int cwid)
{
  int i;
  static int colorwid[]= { Wcolred, Wcolblue, Wcolyellow, Wcolbrown };
  for(i=0;i<4;i++)
   if (cwid==colorwid[i])
   {
     color= i+1;
     gfx_set_state(cwid,1);
   }
   else
     gfx_set_state(colorwid[i],0);
}

static int pixcoor(int x,int y,int *Rx,int *Ry)
{
  if (x<pixOFFx) return 1;
  if (x>=pixOFFx+pixiw) return 1;
  if (y<pixOFFy) return 1;
  if (y>=pixOFFy+pixih) return 1;
  *Rx= (x-pixOFFx)/(PIXSIZ+1);
  *Ry= (y-pixOFFy)/(PIXSIZ+1);
  return 0;
}

static void putpat(int pn)
{
  char *K; int w,h,i; char buf[200];
  if (!pats || pn<0 || pn>=npats) return ;
  K= pats[pn];
  if (getwh(K,&w,&h)) return ;
  setsize(w,h,0);
  for(i=0;i<w*h;i++) pix[i]= K[4+i]-'0';
  render_all_pix();
  sprintf(buf,"%d/%d",pn+1,npats);
  gfx_set_text(Wpage,buf);
  cpat= pn;
  sprintf(buf,"%d", w); gfx_set_text(Wcols, buf);
  sprintf(buf,"%d", h); gfx_set_text(Wrows, buf);
  patmodified= 0;
}

static void getpat()
{
  char *K; int i;
  if (!patmodified) return ;
  if (cpat<0 || cpat>=npats) return ;
  K= newpat(pW,pH);
  for(i=0;i<pW*pH;i++) K[4+i]= pix[i]+'0';
  free(pats[cpat]);
  pats[cpat]= K;
  patmodified= 0;
}

static void addpat()
{
  int A,i;
  getpat();
  if (npats==spats-1)
  {
    spats+= spats/2 + 10;
    pats= realloc(pats, sizeof(pats[0])*spats);
  }
  A= cpat+1; // A= cpat inserts one BEFORE cpat
  for(i=npats;i>A;i--)
   pats[i]= pats[i-1];
  pats[A]= newpat(pW,pH);
  modified= 1;
  npats++;
  putpat(A);
}

static void delpat()
{
  int i;
  free(pats[cpat]);
  for(i=cpat;i<npats-1;i++)
    pats[i]= pats[i+1];
  npats--;
  putpat(cpat); 
  modified= 1;
}

static char *textfiles[]= { "Pattern List Files (*.txt)", NULL };
static char *cfiles[]= { "C Files (*.c)", NULL };
static char *pngfiles[]= { "PNG Images (*.png)", NULL };

static int dosave()
{
  char *fn;
  if (!modified) return 0;
  getpat();
  if (!cfn)
  {
    fn= gfx_select_file(1,NULL,textfiles);
    if (!fn) return 1;
  }
  else
  {
    fn= cfn;
  }
  pats[npats]= NULL;
  if (save_file(fn,pats))
  {
    gfx_ask("Error", "!OK", "There was an error saving file %s",fn);
    return 1;
  }
  cfn= fn;
  patmodified= modified= 0;
  return 0;
}

static int ask_for_save()
{
  int R;
  if (!modified) return 0; 
  getpat();
  if (cfn)
    R= gfx_ask("Save File?", "!Save!Proceed No Save!Cancel",
            "Would you like to save file %s?",cfn);
  else
    R= gfx_ask("Save File?", "!Save!Proceed No Save!Cancel",
            "Would you like to save modified contents?");
  switch(R)
  {
    case 1: return 0;
    case 2: return 1;
  }
  return dosave(); 
}

static void doload()
{
  char *fn;
  char **lines;
  int n;
  
  fn= gfx_select_file(0,cfn,textfiles);
  if (!fn) return ;
  lines= load_file(fn);
  if (!lines)
  {
    gfx_ask("Error", "!OK", "There was an error loading %s",fn);
    return ;
  }
  for(n=0;lines[n];n++)  ;
  if (!n)
  {
    gfx_ask("Error", "!OK", "Empty file %s",fn);
    return ;
  }
  destroy_pats();
  pats= lines;
  npats= n;
  spats= n+1;
  modified= patmodified= 0;
  putpat(0);
  cfn= fn;
}


static void exportC()
{
  char *fn;
  getpat();
  fn= gfx_select_file(1,NULL,cfiles);
  if (!fn) return ; 
  pats[npats]= NULL;
  if (export_c_file(fn,pats))
    gfx_ask("Error","!OK","Error writing %s",fn);
}

static void exportPNG()
{
  char *fn;
  char *mws; int mw;
  getpat();
  fn= gfx_select_file(1,NULL,pngfiles);
  if (!fn) return ; 
  pats[npats]= NULL;
  mws= gfx_get_text(Wmaxw);
  if (!mws) mws= "0";
  mw= atoi(mws);
  if (mw<=0) mw= 320;
  if (export_png_file(fn,pats,mw))
    gfx_ask("Error","!OK","Error writing %s",fn);
}

static void put_digit
   (unsigned int *img, int W,int H, int ox, int oy,char *D)
{
  int dW, dH;
  int x,y;
  dW= D[0];
  dH= D[1];
  D+= 2;
  for(y=0;y<dH;y++) for(x=0;x<dW;x++)
    if (D[x+y*dW])
      img[ (x+ox)+(y+oy)*W ]= 0xccaa00;
}

static void export_single_pattern_PNG()
{
  char *fn;
  int x,y,i;
  unsigned int *si, sW, sH;
  char *e;
  fn= gfx_select_file(1, NULL, pngfiles);
  if (!fn) return ;
  sW= pixiw + 24; sH= pixih + 24;
#include "digitout.c"
#undef NPAT
  si= calloc(sW*sH, sizeof(si[0]));
  for(y=0;y<pixih;y++) for(x=0;x<pixiw;x++) 
    si[(24+x) + (24+y)*sW]= image[(x+pixOFFx)+(y+pixOFFy)*iW];
  for(x=0;x<pW;x++)
    put_digit(si,sW,sH, (x+1)*(PIXSIZ+1), 0, PAT[x%10]);
  for(y=0;y<pH;y++)
    put_digit(si,sW,sH, 0, (y+1)*(PIXSIZ+1), PAT[y%10]);
  for(i=0;i<sW*sH;i++) si[i] |= (0xff<<24);
  e= write_png_image(fn,sW,sH,si);
  if (e)
   gfx_ask("Error", "!OK", "Writing image %s failed: %s", fn, e);
  free(si);
}


int main()
{
  int i1, i2,p; char *s1, *s2;
  gfx_init(NULL, iW+10, iH+81, "Bitmap Maker");
  gfx_SCREEN_0();
  gfx_canvas_setsize(Wcan,iW,iH);
  gfx_canvas_setsize(Wsma,32,32);
  setcolor(Wcolred);
  mkinitpats(1);
  putpat(0);

  while(1)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT: if (ask_for_save()) break; goto done;
    case Wcolred: case Wcolblue: case Wcolyellow: case Wcolbrown:
      setcolor(event.type); 
      break;
    case Wresize: s1= gfx_get_text(Wcols); s2= gfx_get_text(Wrows);
      setsize(s1? atoi(s1):0, s2?atoi(s2):0,1);
      break;
    case GFXEV(Wcan,BUTTON_PRESS):
      if (event.button==1) p= color; else p= 0;
      if (pixcoor(event.x, event.y, &i1, &i2)) continue;
      pix[i1+i2*pW]= p;
      render_pix(i1,i2,1);
      patmodified= modified= 1;
      break;
    case Wnext: if (cpat==npats-1) break; getpat(); putpat(cpat+1); break;
    case Wprev: if (cpat==0) break; getpat(); putpat(cpat-1); break;
    case Wadd: addpat(); break;
    case Wdel: delpat(); break;
    case Wclr: memset(pix,0,sizeof(pix)); render_bg(); render_all_pix();
               patmodified= modified= 1; break;
    case Wnew: 
       if (ask_for_save()) break;
       mkinitpats(0);
       putpat(0);
       break;
    case Wload: if (ask_for_save()) break; doload(); break;
    case Wsave: dosave(); break;
    case WexC: exportC(); break;
    case WexPNG: exportPNG(); break;
    case Wsingle: export_single_pattern_PNG(); break;
    }
  }
done:
  gfx_close();
  return 0;
}
